from . import _api

if _api.USED_API == _api.QT_API_PYQT5:
    from PyQt5.QtTest import *
elif _api.USED_API == _api.QT_API_PYQT4:
    from PyQt4.QtTest import *
elif _api.USED_API == _api.QT_API_PYSIDE:
    from PySide.QtTest import *
elif _api.USED_API == _api.QT_API_PYSIDE2:
    from PySide2.QtTest import *

def _QTest_qWaitForWindowExposed(widget, timeout=1000):
    # A Qt5 compatible (probably) QTest.qWaitForWindowExposed(QWidget, int)
    # (mostly copied from qtestsystem.h in qt5/qtbase)
    from AnyQt.QtCore import \
        Qt, QCoreApplication, QEventLoop, QElapsedTimer, QEvent
    window = widget.window()
    timer = QElapsedTimer()
    timer.start()

    # Is widget.testAttribute(Qt.WA_Mapped) a suitable replacement for
    # QWindow.isExposed in Qt5??
    # Not exactly. In Qt5
    # window().testAttribute(Qt.WA_Mapped) == window().windowHandle.isExposed()
    # but both are False if a window is fully obscured by other windows,
    # in Qt4 there is no difference if a window is obscured.
    while not window.testAttribute(Qt.WA_Mapped):
        remaining = timeout - timer.elapsed()
        if remaining <= 0:
            break
        QCoreApplication.processEvents(QEventLoop.AllEvents, remaining)
        QCoreApplication.sendPostedEvents(None, QEvent.DeferredDelete)
        QTest.qSleep(10)

    return window.testAttribute(Qt.WA_Mapped)


def _QTest_qWaitForWindowActive(widget, timeout=1000):
    # A Qt5 compatible (probably) QTest.qWaitForWindowActive(QWidget, int)
    # (mostly copied from qtestsystem.h in qt5/qtbase)
    from AnyQt.QtCore import \
        Qt, QCoreApplication, QEventLoop, QElapsedTimer, QEvent
    window = widget.window()
    timer = QElapsedTimer()
    timer.start()

    while not window.isActiveWindow():
        remaining = timeout - timer.elapsed()
        if remaining <= 0:
            break
        QCoreApplication.processEvents(QEventLoop.AllEvents, remaining)
        QCoreApplication.sendPostedEvents(None, QEvent.DeferredDelete)
        QTest.qSleep(10)

    # See the explanation in qtestsystem.h
    if window.isActiveWindow():
        wait_no = 0
        while window.pos().isNull():
            if wait_no > timeout // 10:
                break
            wait_no += 1
            QTest.qWait(10)

    return window.isActiveWindow()


if _api.USED_API in {_api.QT_API_PYQT4, _api.QT_API_PYSIDE}:
    QTest.qWaitForWindowExposed = _QTest_qWaitForWindowExposed
    QTest.qWaitForWindowActive = _QTest_qWaitForWindowActive

    from AnyQt.QtCore import QObject, QByteArray as _QByteArray

    # not exposed in PyQt4 or PySide. Going by PyQt5 interface
    class QSignalSpy(QObject):
        """
        QSignalSpy(boundsignal)
        """
        def __init__(self, boundsig, **kwargs):
            super(QSignalSpy, self).__init__(**kwargs)
            from AnyQt.QtCore import QEventLoop, QTimer
            self.__boundsig = boundsig
            self.__recorded = recorded = []  # type: List[List[Any]]
            self.__loop = loop = QEventLoop()
            self.__timer = QTimer(self, singleShot=True)
            self.__timer.timeout.connect(self.__loop.quit)

            def record(*args):
                # Record the emitted arguments and quit the loop if running.
                # NOTE: not capturing self from parent scope
                recorded.append(list(args))
                if loop.isRunning():
                    loop.quit()
            # Need to keep reference at least for PyQt4 4.11.4, sip 4.16.9 on
            # python 3.4 (if the signal is emitted during gc collection, and
            # the boundsignal is a QObject.destroyed signal).
            self.__record = record
            boundsig.connect(record)

        def signal(self):
            return _QByteArray(self.__boundsig.signal[1:].encode("latin-1"))

        def isValid(self):
            return True

        def wait(self, timeout=5000):
            count = len(self)
            self.__timer.stop()
            self.__timer.setInterval(timeout)
            self.__timer.start()
            self.__loop.exec_()
            self.__timer.stop()
            return len(self) != count

        def __getitem__(self, index):
            return self.__recorded[index]

        def __setitem__(self, index, value):
            self.__recorded.__setitem__(index, value)

        def __delitem__(self, index):
            self.__recorded.__delitem__(index)

        def __len__(self):
            return len(self.__recorded)

    del QObject


def _QTest_qWaitFor(predicate, timeout=5000):
    # type: (Callable[[], bool], int) -> bool
    # Copied and adapted from Qt
    from AnyQt.QtCore import Qt, QCoreApplication, QEvent, QEventLoop, QDeadlineTimer
    if predicate():
        return True
    remaining = timeout
    deadline = QDeadlineTimer(remaining, Qt.PreciseTimer)
    while True:
        QCoreApplication.processEvents(QEventLoop.AllEvents)
        QCoreApplication.sendPostedEvents(None, QEvent.DeferredDelete)
        remaining = deadline.remainingTime()
        if remaining > 0:
            QTest.qSleep(min(10, remaining))
        if predicate():
            return True
        remaining = deadline.remainingTime()
        if remaining <= 0:
            break
    return predicate()  # Last chance


if not hasattr(QTest, "qWaitFor"):  # Qt < 5.10
    QTest.qWaitFor = _QTest_qWaitFor



import sys
import warnings

from ._api import (
    USED_API, QT_API_PYQT4, QT_API_PYQT5, QT_API_PYSIDE
)


class ImportHookBackport(object):
    """
    A python import hook (PEP-302) substituting Qt4 module imports, replacing
    them with a back compatible shim.
    """
    def __init__(self, whichapi):
        self.whichapi = whichapi

    def find_module(self, name, path=None):
        if USED_API != QT_API_PYQT5:
            return

        toplevel = name.split(".", 1)[0]
        if toplevel == "PyQt4" and self.whichapi == QT_API_PYQT4:
            return self
        elif toplevel == "PySide" and self.whichapi == QT_API_PYSIDE:
            return self
        else:
            return

    def load_module(self, fullname):
        if fullname in sys.modules:
            # don't reload
            return sys.modules[fullname]

        pkgpath = fullname.split(".")
        toplevel = pkgpath[0]
        subpkg = pkgpath[1] if len(pkgpath) > 1 else None

        assert toplevel.lower() == self.whichapi
        backportpkg = "AnyQt._backport"
        if subpkg is not None:
            backportpkg += "." + subpkg
        module = __import__(backportpkg, fromlist=["_"])
        warnings.warn(
            "Loaded module {} as a substitute for {}"
            .format(module.__name__, fullname),
            RuntimeWarning, stacklevel=2,
        )
        sys.modules[fullname] = module
        return module


class ImportHookDeny(object):
    """
    A python import hook (PEP-302) preventing imports of a Qt api.

    Parameters
    ----------
    whichapi : str
        The Qt api whose import should be prevented.
    """
    def __init__(self, whichapi):
        self.whichapi = whichapi

    def find_module(self, name, path=None):
        toplevel = name.split(".")[0]
        if self.whichapi == QT_API_PYQT5 and toplevel == "PyQt5":
            return self
        elif self.whichapi == QT_API_PYQT4 and toplevel == "PyQt4":
            return self
        elif self.whichapi == QT_API_PYSIDE and toplevel == "PySide":
            return self
        else:
            return None

    def load_module(self, fullname):
        raise ImportError(
            "Import of {} is denied.".format(fullname)
        )


def install_backport_hook(api):
    """
    Install a backport import hook for Qt4 api

    Parameters
    ----------
    api : str
        The Qt4 api whose structure should be intercepted
        ('pyqt4' or 'pyside').

    Example
    -------
    >>> install_backport_hook("pyqt4")
    >>> import PyQt4
    Loaded module AnyQt._backport as a substitute for PyQt4

    """
    if api == USED_API:
        raise ValueError

    sys.meta_path.insert(0, ImportHookBackport(api))


def install_deny_hook(api):
    """
    Install a deny import hook for Qt api.

    Parameters
    ----------
    api : str
        The Qt api whose import should be prevented

    Example
    -------
    >>> install_deny_import("pyqt4")
    >>> import PyQt4
    Traceback (most recent call last):...
    ImportError: Import of PyQt4 is denied.

    """
    if api == USED_API:
        raise ValueError

    sys.meta_path.insert(0, ImportHookDeny(api))



from . import _api

# Names imported from Qt4's QtWebKit module
__Qt4_QtWebKit = [
    'QGraphicsWebView',
    'QWebFrame',
    'QWebHitTestResult',
    'QWebInspector',
    'QWebPage',
    'QWebView',
]

if _api.USED_API == _api.QT_API_PYQT5:
    from PyQt5.QtWebKitWidgets import *
elif _api.USED_API == _api.QT_API_PYQT4:
    from PyQt4.QtWebKit import (
        QGraphicsWebView,
        QWebFrame,
        QWebHitTestResult,
        QWebInspector,
        QWebPage,
        QWebView,
    )
elif _api.USED_API == _api.QT_API_PYSIDE:
    from PySide.QtWebKit import (
        QGraphicsWebView,
        QWebFrame,
        QWebHitTestResult,
        QWebInspector,
        QWebPage,
        QWebView,
    )



from . import _api

# Names imported from Qt4's QtWebKit module
__Qt4_QtWebKit = [
    'QWebDatabase',
    'QWebElement',
    'QWebElementCollection',
    'QWebHistory',
    'QWebHistoryInterface',
    'QWebHistoryItem',
    'QWebPluginFactory',
    'QWebSecurityOrigin',
    'QWebSettings',
    'qWebKitMajorVersion',
    'qWebKitMinorVersion',
    'qWebKitVersion'
]

if _api.USED_API == _api.QT_API_PYQT5:
    from PyQt5.QtWebKit import *
elif _api.USED_API == _api.QT_API_PYQT4:
    from PyQt4.QtWebKit import (
        QWebDatabase,
        QWebElement,
        QWebElementCollection,
        QWebHistory,
        QWebHistoryInterface,
        QWebHistoryItem,
        QWebPluginFactory,
        QWebSecurityOrigin,
        QWebSettings,
        qWebKitMajorVersion,
        qWebKitMinorVersion,
        qWebKitVersion
    )
elif _api.USED_API == _api.QT_API_PYSIDE:
    from PySide.QtWebKit import (
        QWebDatabase,
        QWebElement,
        QWebElementCollection,
        QWebHistory,
        QWebHistoryInterface,
        QWebHistoryItem,
        QWebPluginFactory,
        QWebSecurityOrigin,
        QWebSettings,
    )
    try:
        # missing in current PySide 1.2.2
        from PySide.QtWebKit import \
            qWebKitMajorVersion, qWebKitMinorVersion, qWebKitVersion
    except ImportError:
        pass
elif _api.USED_API == _api.QT_API_PYSIDE2:
    from PySide2.QtWebKit import *



from . import _api

if _api.USED_API == _api.QT_API_PYQT5:
    from PyQt5.QtWinExtras import *
elif _api.USED_API == _api.QT_API_PYSIDE2:
    from PySide2.QtWinExtras import *
else:
    raise ImportError
